import json
import pathlib
from typing import List

from ..schema import Example
from .developer import Developer

config_path = pathlib.Path(__file__).parent / ".autom" / "interface_designer.json"
with open(config_path, "r") as f:
    config_data: dict = json.load(f)


class InterfaceDesigner(Developer):
    @property
    def model(self) -> str:
        return config_data["model"]

    @property
    def role(self) -> str:
        return config_data["title"]

    @property
    def job_description(self) -> str:
        return config_data["job-description"]

    @property
    def format_requirements(self) -> List[str]:
        requirements = config_data.get("extra-format-requirements", None)
        if requirements is None:
            requirements = []
        parent_format_requirements = super().format_requirements
        return parent_format_requirements + requirements

    @property
    def guidance_requirements(self) -> List[str]:
        requirements = config_data.get("extra-guidance-requirements", None)
        if requirements is None:
            requirements = []
        parent_guidance_requirements = super().guidance_requirements
        return parent_guidance_requirements + requirements

    @property
    def examples(self) -> List[Example]:
        example_dict_list = config_data.get("examples", None)
        if example_dict_list is None:
            return []
        return [Example(example_data["query"], example_data["response"]) for example_data in config_data["examples"]]



from .developer import Developer
from .interface_designer import InterfaceDesigner
from .api_documentation_architect import APIDocumentationArchitect


__all__ = ["Developer", "InterfaceDesigner", "APIDocumentationArchitect"]



import pathlib


class CodeWriter:
    """
    A utility class for writing code to files.

    Now, `CodeWriter` only supports Python code.
    """
    @staticmethod
    def ensure_init_py(directory: pathlib.Path):
        """
        Recursively ensure that each directory level has an __init__.py file.
        """
        if directory.exists() and not directory.is_dir():
            raise ValueError(f"{directory} exists but is not a directory!")

        if not directory.exists():
            directory.mkdir(parents=True, exist_ok=True)

        init_file = directory / '__init__.py'
        if not init_file.exists():
            init_file.touch()

        # Recurse to parent until reaching the root
        parent = directory.parent
        if parent != directory:  # Stop condition for reaching root
            CodeWriter.ensure_init_py(parent)

    @staticmethod
    def write_code(file_path: pathlib.Path, code: str):
        """
        Write code to the given file_path, ensuring directories and __init__.py files exist.
        """
        CodeWriter.ensure_init_py(file_path.parent)

        with open(file_path, "a") as f:
            f.write(code)


__all__ = ['CodeWriter']



from typing import Optional


class LLMResponse:
    """
    `LLMResponse` stores information parsed from LLM's response
    """

    def __init__(self, code: Optional[str] = None, explanation: Optional[str] = None):
        self.code = code
        self.explanation = explanation

    def __str__(self):
        code_brief = self.code[:20] + "..." if len(self.code) > 20 else self.code
        explanation_brief = self.explanation[:20] + "..." if len(self.explanation) > 20 else self.explanation
        return f"LLMResponse(code={code_brief}, explanation={explanation_brief})"

    def __repr__(self):
        return self.__str__()


class Example:
    """
    `Example` stores an example for the developer to learn how to do the task
    """
    def __init__(self, query: str, response: str):
        self.query = query
        self.response = response

    def __str__(self):
        query_brief = self.query[:20] + "..." if len(self.query) > 20 else self.query
        response_brief = self.response[:20] + "..." if len(self.response) > 20 else self.response
        return f"Example(query={query_brief}, response={response_brief})"

    def __repr__(self):
        return self.__str__()



import json
import pathlib
from typing import List

from langchain import PromptTemplate
from langchain.chat_models import ChatOpenAI
from gembox.format_utils import to_sorted_list_str
from gembox.collection_utils import is_empty_collection
from tenacity import retry, stop_after_attempt, retry_if_exception_type

from ..schema import LLMResponse, Example

config_path = pathlib.Path(__file__).parent / ".autom" / "developer.json"
with open(config_path, "r") as f:
    developer_config_data: dict = json.load(f)


class Developer:
    """
    This is the abstract class for a GPT-based developer.

    A well-defined developer should be defined by:
    - role: The role the developer plays.
    - work_output_requirement: The type of work result the developer needs to produce.

    The developer always should convey the following as their working result:
    - code: The code that can accomplish the user's task.
    - explanation: The explanation of the code, which can be used to explain the code to the user.
    """

    def __init__(self):
        self._llm = ChatOpenAI(model=self.model)

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(Exception))
    def serve(self, user_msg: str):
        """
        Serve the user's message and set up the work result.
        """
        # 1. Query the llm
        llm_resp_text = self._query_llm(user_msg)

        # 2. Parse the llm response
        llm_resp = self._parse_llm_resp(llm_resp_text)

        # 3. return the result
        return llm_resp

    def _query_llm(self, user_msg: str) -> str:
        return self.llm.predict(self.template.format(user_message=user_msg))

    def _parse_llm_resp(self, llm_resp_text: str) -> LLMResponse:
        code = self._get_code_from_llm_resp(llm_resp_text)
        explanation = self._get_explanation_from_llm_resp(llm_resp_text)
        return LLMResponse(code=code, explanation=explanation)

    def _get_code_from_llm_resp(self, llm_resp_text: str) -> str:
        """parse the code from the llm response"""
        code_start, code_end = "MAGIC$$$", "$$$"
        code = llm_resp_text.split(code_start)[1].split(code_end)[0].strip()
        return code

    def _get_explanation_from_llm_resp(self, llm_resp_text: str) -> str:
        """parse the explanation from the llm response"""
        explanation_start, explanation_end = "MAGIC%%%", "%%%"
        explanation = llm_resp_text.split(explanation_start)[1].split(explanation_end)[0].strip()
        return explanation

    @property
    def template(self) -> PromptTemplate:
        template_str = ""
        if self.job_description is not None:
            template_str += (self.job_description + "\n")

        requirements = self.requirements
        if not is_empty_collection(requirements):
            template_str += "Requirements:\n"
            template_str += to_sorted_list_str(list(requirements))

        # add examples
        examples = self.examples
        if not is_empty_collection(examples):
            template_str += "\n"
            for i, example in enumerate(examples):
                template_str += f"e.g.{i + 1}\n[Query]:\n{example.query}\n[Response]:\n{example.response}\n"
            template_str += "Now, it's your turn.\n"

        template_str += "[Query]:\n{user_message}\n[Response]:"

        return PromptTemplate.from_template(template_str)

    @property
    def model(self) -> str:
        return developer_config_data["model"]

    @property
    def role(self) -> str:
        return developer_config_data["title"]

    @property
    def job_description(self) -> str:
        # if no key, then return None
        return developer_config_data["job-description"]

    @property
    def guidance_requirements(self) -> List[str]:
        requirements = developer_config_data.get("extra-guidance-requirements", None)
        if requirements is None:
            return []
        return requirements

    @property
    def format_requirements(self) -> List[str]:
        requirements = developer_config_data.get("extra-format-requirements", None)
        if requirements is None:
            return []
        return requirements

    @property
    def examples(self) -> List[Example]:
        example_dict_list = developer_config_data.get("examples", None)
        if example_dict_list is None:
            return []
        return [Example(example_data["query"], example_data["response"]) for example_data in
                developer_config_data["examples"]]

    @property
    def system_message(self) -> str:
        return f"You are a {self.role} to translate user's needs into high quality code."

    @property
    def requirements(self) -> List[str]:
        """returns the requirement of the work output, including the requirements for guidance and formatting"""
        return self.guidance_requirements + self.format_requirements

    @property
    def llm(self):
        """get the llm instance"""
        return self._llm


__all__ = ['Developer']



from setuptools import setup, find_packages


def read_requirements():
    with open('requirements.txt', 'r') as f:
        requirements = f.readlines()
    return [req.strip() for req in requirements if req.strip() and not req.startswith('#')]


setup(
    name='tootle',
    version='0.0.1.1',
    include_package_data=True,
    packages=find_packages(exclude=['test', 'test.*']),
    package_data={
        'tootle.dev_bot': ['.autom/*.json'],
    },
    install_requires=read_requirements(),
    author='',
    author_email='',
    description='',
    long_description=open('./README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
)

# To build:
# python setup.py sdist bdist_wheel
# To upload to pypi:
# twine upload dist/*
